@@ -10,8 +10,8 @@
 use std::sync::Arc;
 
 use crate::input::proto::substrait;
-use crate::output::data_type;
 use crate::output::diagnostic;
+use crate::output::type_system::data;
 use crate::parse::context;
 use crate::parse::expressions;
 use crate::parse::expressions::literals;
@@ -27,7 +27,7 @@ struct SourceInfo {
     pub name: String,
 
     /// The schema of the data, if not context-sensitive.
-    pub data_type: Option<Arc<data_type::DataType>>,
+    pub data_type: Option<data::Type>,
 }
 
 /// Parse virtual table.
@@ -35,7 +35,7 @@ fn parse_virtual_table(
     x: &substrait::read_rel::VirtualTable,
     y: &mut context::Context,
 ) -> diagnostic::Result<SourceInfo> {
-    let mut data_type: Arc<data_type::DataType> = Arc::default();
+    let mut data_type: data::Type = Arc::default();
 
     // Parse rows, ensuring that they all have the same type.
     proto_repeated_field!(x, y, values, |x, y| {
@@ -116,6 +116,9 @@ fn parse_file_format(
         substrait::read_rel::local_files::file_or_files::FileFormat::Orc(_) => {
             Some(String::from("Orc"))
         }
+        substrait::read_rel::local_files::file_or_files::FileFormat::Dwrf(_) => {
+            Some(String::from("Dwrf"))
+        }
         substrait::read_rel::local_files::file_or_files::FileFormat::Extension(x) => {
             extensions::advanced::parse_functional_any(x, y)?;
             None
@@ -334,7 +337,7 @@ pub fn parse_read_rel(x: &substrait::ReadRel, y: &mut context::Context) -> diagn
     y.set_schema(schema.clone());
 
     // Handle filter.
-    proto_boxed_field!(x, y, filter, expressions::parse_predicate);
+    let filter = proto_boxed_field!(x, y, filter, expressions::parse_predicate);
 
     // Handle projection.
     if x.projection.is_some() {
@@ -353,6 +356,17 @@ pub fn parse_read_rel(x: &substrait::ReadRel, y: &mut context::Context) -> diagn
         (true, true) => describe!(y, Relation, "Filtered partial read from {}", source.name),
     }
 
+    // Add filter summary.
+    if let (Some(filter_node), Some(filter_expr)) = filter {
+        let nullable = filter_node.data_type().nullable();
+        summary!(
+            y,
+            "This relation discards all rows for which the expression {} yields {}.",
+            filter_expr,
+            if nullable { "false or null" } else { "false" }
+        );
+    }
+
     // Handle the common field.
     handle_rel_common!(x, y);
 
