@@ -15,7 +15,6 @@ use crate::input::traits::InputNode;
 use crate::input::traits::ProtoEnum;
 use crate::input::yaml;
 use crate::output::diagnostic;
-use crate::output::extension;
 use crate::output::parse_result;
 use crate::output::path;
 use crate::output::primitive_data;
@@ -585,55 +584,34 @@ pub fn parse_proto<T, F, B>(
     root_parser: F,
     state: &mut context::State,
     config: &config::Config,
-) -> parse_result::ParseResult
+) -> diagnostic::Result<parse_result::ParseResult>
 where
     T: prost::Message + InputNode + Default,
     F: FnOnce(&T, &mut context::Context) -> diagnostic::Result<()>,
     B: prost::bytes::Buf,
 {
-    match T::decode(buffer) {
-        Err(err) => {
-            // Create a minimal root node with just the decode error
-            // diagnostic.
-            let mut root = T::type_to_node();
+    // Run protobuf deserialization.
+    let input = T::decode(buffer).map_err(|e| ecause!(ProtoParseFailed, e))?;
 
-            // Create a root context for it.
-            let mut context = context::Context::new(root_name, &mut root, state, config);
+    // Create the root node.
+    let mut root = input.data_to_node();
 
-            // Push the diagnostic using the context.
-            context.push_diagnostic(diagnostic::RawDiagnostic {
-                cause: ecause!(ProtoParseFailed, err),
-                level: diagnostic::Level::Error,
-                path: path::PathBuf {
-                    root: root_name,
-                    elements: vec![],
-                },
-            });
+    // Create the root context.
+    let mut context = context::Context::new(root_name, &mut root, state, config);
 
-            parse_result::ParseResult { root }
-        }
-        Ok(input) => {
-            // Create the root node.
-            let mut root = input.data_to_node();
+    // Call the provided parser function.
+    let success = root_parser(&input, &mut context)
+        .map_err(|cause| {
+            diagnostic!(&mut context, Error, cause);
+        })
+        .is_ok();
 
-            // Create the root context.
-            let mut context = context::Context::new(root_name, &mut root, state, config);
+    // Handle any fields not handled by the provided parse function.
+    // Only generate a warning diagnostic for unhandled children if the
+    // parse function succeeded.
+    handle_unknown_children(&input, &mut context, success);
 
-            // Call the provided parser function.
-            let success = root_parser(&input, &mut context)
-                .map_err(|cause| {
-                    diagnostic!(&mut context, Error, cause);
-                })
-                .is_ok();
-
-            // Handle any fields not handled by the provided parse function.
-            // Only generate a warning diagnostic for unhandled children if the
-            // parse function succeeded.
-            handle_unknown_children(&input, &mut context, success);
-
-            parse_result::ParseResult { root }
-        }
-    }
+    Ok(parse_result::ParseResult { root })
 }
 
 //=============================================================================
@@ -648,6 +626,9 @@ macro_rules! yaml_field {
     ($input:expr, $context:expr, $field:expr, $parser:expr) => {
         crate::parse::traversal::push_yaml_field($input, $context, $field, false, $parser)
     };
+    ($input:expr, $context:expr, $field:expr, $parser:expr, $($args:expr),*) => {
+        yaml_field!($input, $context, $field, |x, y| $parser(x, y, $($args),*))
+    };
 }
 
 /// Parse and push an optional YAML field.
@@ -693,6 +674,9 @@ macro_rules! yaml_required_field {
     ($input:expr, $context:expr, $field:expr, $parser:expr) => {
         crate::parse::traversal::push_yaml_required_field($input, $context, $field, false, $parser)
     };
+    ($input:expr, $context:expr, $field:expr, $parser:expr, $($args:expr),*) => {
+        yaml_required_field!($input, $context, $field, |x, y| $parser(x, y, $($args),*))
+    };
 }
 
 /// Parse and push a required field of a YAML object. If the field does not
@@ -738,6 +722,9 @@ macro_rules! yaml_array {
     ($input:expr, $context:expr, $parser:expr, $min_size:expr) => {
         crate::parse::traversal::push_yaml_array($input, $context, $min_size, false, $parser)
     };
+    ($input:expr, $context:expr, $parser:expr, $min_size:expr, $($args:expr),*) => {
+        yaml_array!($input, $context, |x, y| $parser(x, y, $($args),*), $min_size)
+    };
 }
 
 /// Convenience/shorthand macro for parsing a YAML array that must have at
@@ -749,6 +736,9 @@ macro_rules! yaml_required_array {
     ($input:expr, $context:expr, $parser:expr) => {
         yaml_array!($input, $context, $parser, 1)
     };
+    ($input:expr, $context:expr, $parser:expr, $($args:expr),*) => {
+        yaml_array!($input, $context, $parser, 1, $($args:expr),*)
+    };
 }
 
 /// Parse and push an optional YAML array element.
@@ -844,6 +834,9 @@ macro_rules! yaml_repeated_field {
             $input, $context, $field, false, $min_size, false, $parser,
         )
     };
+    ($input:expr, $context:expr, $field:expr, $parser:expr, $min_size:expr, $($args:expr),*) => {
+        yaml_repeated_field!($input, $context, $field, |x, y| $parser(x, y, $($args),*), $min_size)
+    };
 }
 
 /// Shorthand for fields that must be arrays.
@@ -859,6 +852,9 @@ macro_rules! yaml_required_repeated_field {
             $input, $context, $field, true, $min_size, false, $parser,
         )
     };
+    ($input:expr, $context:expr, $field:expr, $parser:expr, $min_size:expr, $($args:expr),*) => {
+        yaml_required_repeated_field!($input, $context, $field, |x, y| $parser(x, y, $($args),*), $min_size)
+    };
 }
 
 /// Parse and push a complete YAML array. If a required element does not exist,
@@ -992,14 +988,39 @@ pub mod yaml_primitive_parsers {
 }
 
 //=============================================================================
-// YAML root handling
+// URI resolution and YAML root handling
 //=============================================================================
 
-/// Attempts to resolve a URI.
-fn resolve_uri(
-    uri: &str,
-    context: &mut context::Context,
-) -> diagnostic::Result<config::BinaryData> {
+/// Worker for resolve_uri().
+fn resolve_uri(uri: &str, context: &mut context::Context) -> Option<config::BinaryData> {
+    // Check for cyclic dependencies.
+    let uri_stack = context.uri_stack();
+    if let Some((index, _)) = uri_stack.iter().enumerate().find(|(_, x)| &x[..] == uri) {
+        let cycle = itertools::Itertools::intersperse(
+            uri_stack
+                .iter()
+                .map(|x| &x[index..])
+                .chain(std::iter::once(uri)),
+            " -> ",
+        )
+        .collect::<String>();
+        diagnostic!(context, Error, YamlCyclicDependency, "{cycle}");
+        return None;
+    }
+
+    // Check for recursion limit.
+    if let Some(max_depth) = context.config.max_uri_resolution_depth {
+        if context.uri_stack().len() >= max_depth {
+            diagnostic!(
+                context,
+                Warning,
+                YamlResolutionDisabled,
+                "configured recursion limit for URI resolution has been reached"
+            );
+            return None;
+        }
+    }
+
     // Apply yaml_uri_overrides configuration.
     let remapped_uri = context
         .config
@@ -1018,10 +1039,13 @@ fn resolve_uri(
     let remapped_uri = if let Some(remapped_uri) = remapped_uri {
         remapped_uri.to_owned()
     } else {
-        return Err(cause!(
+        diagnostic!(
+            context,
+            Warning,
             YamlResolutionDisabled,
             "YAML resolution for {uri} was disabled"
-        ));
+        );
+        return None;
     };
     if is_remapped {
         diagnostic!(context, Info, Yaml, "URI was remapped to {remapped_uri}");
@@ -1029,92 +1053,155 @@ fn resolve_uri(
 
     // If a custom download function is specified, use it to resolve.
     if let Some(ref resolver) = context.config.uri_resolver {
-        return resolver(&remapped_uri)
-            .map_err(|x| ecause!(YamlResolutionFailed, x.as_ref().to_string()));
+        return match resolver(&remapped_uri) {
+            Ok(x) => Some(x),
+            Err(e) => {
+                diagnostic!(context, Warning, YamlResolutionFailed, "{e}");
+                None
+            }
+        };
     }
 
     // Parse as a URL.
     let url = match url::Url::parse(&remapped_uri) {
         Ok(url) => url,
         Err(e) => {
-            return Err(if is_remapped {
-                cause!(
+            if is_remapped {
+                diagnostic!(
+                    context,
+                    Warning,
                     YamlResolutionFailed,
                     "configured URI remapping ({remapped_uri}) did not parse as URL: {e}"
                 )
             } else {
-                cause!(
+                diagnostic!(
+                    context,
+                    Warning,
                     YamlResolutionFailed,
                     "failed to parse {remapped_uri} as URL: {e}"
                 )
-            });
+            };
+            return None;
         }
     };
 
     // Reject anything that isn't file://-based.
     if url.scheme() != "file" {
-        return Err(if is_remapped {
-            cause!(
+        if is_remapped {
+            diagnostic!(
+                context,
+                Warning,
                 YamlResolutionFailed,
                 "configured URI remapping ({remapped_uri}) does not use file:// scheme"
             )
         } else {
-            cause!(YamlResolutionFailed, "URI does not use file:// scheme")
-        });
+            diagnostic!(
+                context,
+                Warning,
+                YamlResolutionFailed,
+                "URI does not use file:// scheme"
+            )
+        };
+        return None;
     }
 
     // Convert to path.
-    let path = match url.to_file_path() {
-        Ok(path) => path,
-        Err(_) => {
-            return Err(if is_remapped {
-                cause!(
+    let path =
+        match url.to_file_path() {
+            Ok(path) => path,
+            Err(_) => {
+                if is_remapped {
+                    diagnostic!(context, Warning,
                     YamlResolutionFailed,
                     "configured URI remapping ({remapped_uri}) could not be converted to file path"
                 )
-            } else {
-                cause!(
-                    YamlResolutionFailed,
-                    "URI could not be converted to file path"
-                )
-            });
-        }
-    };
+                } else {
+                    diagnostic!(
+                        context,
+                        Warning,
+                        YamlResolutionFailed,
+                        "URI could not be converted to file path"
+                    )
+                };
+                return None;
+            }
+        };
 
     // Read the file.
-    std::fs::read(path)
-        .map_err(|e| {
+    match std::fs::read(path) {
+        Ok(data) => Some(Box::new(data)),
+        Err(e) => {
             if is_remapped {
-                cause!(
+                diagnostic!(
+                    context,
+                    Warning,
                     YamlResolutionFailed,
                     "failed to file remapping for URI ({remapped_uri}): {e}"
-                )
+                );
             } else {
-                ecause!(YamlResolutionFailed, e)
+                ediagnostic!(context, Warning, YamlResolutionFailed, e);
             }
-        })
-        .map(|d| -> config::BinaryData { Box::new(d) })
+            None
+        }
+    }
 }
 
-/// Resolves a URI to a YAML file, parses the YAML syntax, and optionally
-/// validates it using the given JSON schema.
-fn load_yaml(
+/// Attempts to resolve a URI. If resolution fails or is disabled, None is
+/// returned and an appropriate diagnostic may be emitted. This function
+/// includes detection and prevention of recursive resolution. The reader
+/// function will first be called to convert the binary data from the
+/// resolution to a traversable tree (something that satisfied InputNode),
+/// then the parser will be called on the root of that tree.
+pub fn parse_uri<FP, FR, TF, TR>(
     uri: &str,
     context: &mut context::Context,
+    reader: FR,
+    parser: FP,
+) -> OptionalResult<TR>
+where
+    TF: InputNode,
+    FR: FnOnce(config::BinaryData, &mut context::Context) -> Option<TF>,
+    FP: FnOnce(&TF, &mut context::Context) -> diagnostic::Result<TR>,
+{
+    // Try resolving the URI.
+    if let Some(data) = resolve_uri(uri, context) {
+        // Parse the flat file to a traversable tree.
+        if let Some(root) = reader(data, context) {
+            // Update recursion stack.
+            context.uri_stack().push(uri.to_string());
+
+            // Defer to the provided parser to handle parsing the resolved
+            // data.
+            let (field_output, result) = push_child(
+                context,
+                &root,
+                path::PathElement::Field("data".to_string()),
+                false,
+                parser,
+            );
+
+            // Revert recursion stack update.
+            context.uri_stack().pop();
+
+            // Replace node type to make clear what the child node we just
+            // added signifies.
+            context.replace_node_type(tree::NodeType::ResolvedUri(uri.to_string()));
+
+            // Success, at least to the point that a tree was formed. The
+            // tree itself might still be invalid.
+            return (Some(field_output), result);
+        }
+    }
+
+    (None, None)
+}
+
+/// Read function for YAML files, to be used with [parse_uri()].
+pub fn read_yaml(
+    binary_data: config::BinaryData,
+    context: &mut context::Context,
     schema: Option<&jsonschema::JSONSchema>,
 ) -> Option<yaml::Value> {
-    // Try to resolve the YAML file. Note that failure to resolve is a warning,
-    // not an error; it means the plan isn't valid in the current environment,
-    // but it might still be valid in another one, in particular for consumers
-    // that don't need to be able to resolve the YAML files to use the plan.
-    let binary_data = match resolve_uri(uri, context) {
-        Err(e) => {
-            diagnostic!(context, Warning, e);
-            return None;
-        }
-        Ok(x) => x,
-    };
-
     // Parse as UTF-8.
     let string_data = match std::str::from_utf8(binary_data.as_ref().as_ref()) {
         Err(e) => {
@@ -1125,33 +1212,12 @@ fn load_yaml(
     };
 
     // Parse as YAML.
-    let yaml_data = match yaml_rust::YamlLoader::load_from_str(string_data) {
+    let yaml_data = match serde_yaml::from_str::<serde_yaml::Value>(string_data) {
         Err(e) => {
             ediagnostic!(context, Error, YamlParseFailed, e);
             return None;
         }
-        Ok(x) => {
-            if x.len() > 1 {
-                diagnostic!(
-                    context,
-                    Warning,
-                    YamlParseFailed,
-                    "YAML file contains multiple documents; ignoring all but the first"
-                );
-            }
-            match x.into_iter().next() {
-                None => {
-                    diagnostic!(
-                        context,
-                        Error,
-                        YamlParseFailed,
-                        "YAML file contains zero documents"
-                    );
-                    return None;
-                }
-                Some(x) => x,
-            }
-        }
+        Ok(x) => x,
     };
 
     // Convert to JSON DOM.
@@ -1176,76 +1242,202 @@ fn load_yaml(
     Some(json_data)
 }
 
-/// Attempt to load and parse a YAML file using the given root parse function,
-/// initial state, and configuration.
-pub fn parse_yaml<TS, FP>(
-    uri: TS,
-    context: &mut context::Context,
-    schema: Option<&jsonschema::JSONSchema>,
-    parser: FP,
-) -> Arc<extension::YamlInfo>
-where
-    TS: AsRef<str>,
-    FP: Fn(&yaml::Value, &mut context::Context) -> diagnostic::Result<()>,
-{
-    let uri = uri.as_ref();
-    let uri_reference = extension::NamedReference::new(Some(uri), context.parent_path_buf());
+//=============================================================================
+// ANTLR syntax tree node handling
+//=============================================================================
 
-    // Resolve the YAML file.
-    let yaml_info = Arc::new(if let Some(root_input) = load_yaml(uri, context, schema) {
-        // Create an empty YamlData object.
-        *context.yaml_data_opt() = Some(extension::YamlData::new(uri_reference));
+/// Wrapper type to satisfy push_child()'s InputNode trait bound on the input
+/// node type.
+struct AntlrContextWrapper<'a, T>(&'a T);
 
-        // Create the node for the YAML data root.
-        let mut root_output = root_input.data_to_node();
+impl<'a, T> InputNode for AntlrContextWrapper<'a, T> {
+    fn type_to_node() -> tree::Node {
+        tree::NodeType::AstNode.into()
+    }
 
-        // Create the path element for referring to the YAML data root.
-        let path_element = path::PathElement::Field("data".to_string());
+    fn data_to_node(&self) -> tree::Node {
+        tree::NodeType::AstNode.into()
+    }
 
-        // Create the context for the YAML data root.
-        let mut root_context = context.child(&mut root_output, path_element.clone());
+    fn oneof_variant(&self) -> Option<&'static str> {
+        None
+    }
 
-        // Create a PathBuf for the root node.
-        let root_path = root_context.path_buf();
+    fn parse_unknown(&self, _: &mut context::Context<'_>) -> bool {
+        false
+    }
+}
 
-        // Call the provided root parser.
-        let success = parser(&root_input, &mut root_context)
-            .map_err(|cause| {
-                diagnostic!(&mut root_context, Error, cause);
-            })
-            .is_ok();
+/// Convenience/shorthand macro for traversing into a syntax tree node by node.
+macro_rules! antlr_child {
+    ($input:expr, $context:expr, $field:ident, $analyzer:expr) => {
+        antlr_child!($input, $context, $field, 0, $analyzer)
+    };
+    ($input:expr, $context:expr, $field:ident, $index:expr, $analyzer:expr) => {
+        crate::parse::traversal::push_antlr_child(
+            $context,
+            $input,
+            $index,
+            stringify!($field),
+            $analyzer,
+        )
+    };
+    ($input:expr, $context:expr, $field:ident, $index:expr, $analyzer:expr, $($args:expr),*) => {
+        antlr_child!($input, $context, $field, $index, |x, y| $analyzer(x, y, $($args),*))
+    };
+}
 
-        // Handle any fields not handled by the provided parse function.
-        handle_unknown_children(&root_input, &mut root_context, success);
+/// Parse and push a child of an ANTLR syntax tree node.
+pub fn push_antlr_child<'input, TP, TC, TR, FA>(
+    context: &mut context::Context,
+    parent: &TP,
+    index: usize,
+    field: &'static str,
+    analyzer: FA,
+) -> OptionalResult<TR>
+where
+    TP: antlr_rust::parser_rule_context::ParserRuleContext<'input>,
+    FA: FnOnce(&TC, &mut context::Context) -> diagnostic::Result<TR>,
+    TC: antlr_rust::parser_rule_context::ParserRuleContext<'input, TF = TP::TF, Ctx = TP::Ctx>
+        + 'input,
+{
+    if let Some(child) = parent.child_of_type::<TC>(index) {
+        let (field_output, result) = push_child(
+            context,
+            &AntlrContextWrapper(child.as_ref()),
+            path::PathElement::Field(field.to_string()),
+            false,
+            |x: &AntlrContextWrapper<TC>, y| analyzer(x.0, y),
+        );
+        (Some(field_output), result)
+    } else {
+        (None, None)
+    }
+}
 
-        // Push and return the completed node.
-        let root_output = Arc::new(root_output);
-        context.push(tree::NodeData::Child(tree::Child {
-            path_element,
-            node: root_output.clone(),
-            recognized: true,
-        }));
+/// Convenience/shorthand macro for traversing into a syntax tree node by node.
+/// Contrary to antlr_child! and most other traversal macros, this does NOT
+/// make a child node in the resulting tree. It can be used to hide unobvious
+/// grammar constructs, such rules related to avoiding left recursion.
+macro_rules! antlr_hidden_child {
+    ($input:expr, $context:expr, $analyzer:expr) => {
+        antlr_hidden_child!($input, $context, 0, $analyzer)
+    };
+    ($input:expr, $context:expr, $index:expr, $analyzer:expr) => {
+        crate::parse::traversal::push_antlr_hidden_child(
+            $context,
+            $input,
+            $index,
+            $analyzer,
+        )
+    };
+    ($input:expr, $context:expr, $index:expr, $analyzer:expr, $($args:expr),*) => {
+        antlr_hidden_child!($input, $context, $index, |x, y| $analyzer(x, y, $($args),*))
+    };
+}
 
-        // Take the constructed YAML data object from the context.
-        let mut yaml_data = context.yaml_data_opt().take().unwrap();
-
-        // Configure the reference to the root node in the YamlData object.
-        yaml_data.data.path = root_path;
-        yaml_data.data.node = root_output;
+/// Parse and push a child of an ANTLR syntax tree node, without making a
+/// corresponding child node in the output tree.
+pub fn push_antlr_hidden_child<'input, TP, TC, TR, FA>(
+    context: &mut context::Context,
+    parent: &TP,
+    index: usize,
+    analyzer: FA,
+) -> Option<TR>
+where
+    TP: antlr_rust::parser_rule_context::ParserRuleContext<'input>,
+    FA: FnOnce(&TC, &mut context::Context) -> diagnostic::Result<TR>,
+    TC: antlr_rust::parser_rule_context::ParserRuleContext<'input, TF = TP::TF, Ctx = TP::Ctx>
+        + 'input,
+{
+    parent.child_of_type::<TC>(index).and_then(|child| {
+        analyzer(child.as_ref(), context)
+            .map_err(|cause| {
+                diagnostic!(context, Error, cause);
+            })
+            .ok()
+    })
+}
 
-        // Wrap the completed YAML data object in an Arc.
-        let yaml_data = Arc::new(yaml_data);
+/// This does more or less the opposite of pushing a hidden child: it creates a
+/// child node in the output tree without traversing deeper into the input
+/// tree. It can be used to hide unobvious grammar constructs, such rules
+/// related to avoiding left recursion.
+macro_rules! antlr_recurse {
+    ($input:expr, $context:expr, $field:ident, $analyzer:expr) => {
+        crate::parse::traversal::push_antlr_recurse(
+            $context,
+            $input,
+            stringify!($field),
+            $analyzer,
+        )
+    };
+    ($input:expr, $context:expr, $field:ident, $analyzer:expr, $($args:expr),*) => {
+        antlr_recurse!($input, $context, $field, |x, y| $analyzer(x, y, $($args),*))
+    };
+}
 
-        // The node type will have been set as if this is a normal string
-        // primitive. We want extra information though, namely the contents of
-        // the YAML file. So we change the node type.
-        context.replace_node_type(tree::NodeType::YamlReference(yaml_data.clone()));
+/// Parse and push a child of an ANTLR syntax tree node.
+pub fn push_antlr_recurse<'input, TP, TR, FA>(
+    context: &mut context::Context,
+    parent: &TP,
+    field: &'static str,
+    analyzer: FA,
+) -> RequiredResult<TR>
+where
+    TP: antlr_rust::parser_rule_context::ParserRuleContext<'input>,
+    FA: FnOnce(&TP, &mut context::Context) -> diagnostic::Result<TR>,
+{
+    push_child(
+        context,
+        &AntlrContextWrapper(parent),
+        path::PathElement::Field(field.to_string()),
+        false,
+        |x: &AntlrContextWrapper<TP>, y| analyzer(x.0, y),
+    )
+}
 
-        // Construct the YAML information object.
-        extension::YamlInfo::Resolved(yaml_data)
-    } else {
-        extension::YamlInfo::Unresolved(uri_reference)
-    });
+/// Convenience/shorthand macro for traversing into all children of a certain
+/// type in a syntax tree.
+macro_rules! antlr_children {
+    ($input:expr, $context:expr, $rule:ident, $analyzer:expr) => {
+        crate::parse::traversal::push_antlr_children(
+            $context,
+            $input,
+            stringify!($rule),
+            $analyzer,
+        )
+    };
+    ($input:expr, $context:expr, $rule:ident, $analyzer:expr, $($args:expr),*) => {
+        antlr_children!($input, $context, $rule, |x, y| $analyzer(x, y, $($args),*))
+    };
+}
 
-    yaml_info
+/// Parse and push a child of an ANTLR syntax tree node.
+pub fn push_antlr_children<'input, TP, TC, TR, FA>(
+    context: &mut context::Context,
+    parent: &TP,
+    field: &'static str,
+    mut analyzer: FA,
+) -> RepeatedResult<TR>
+where
+    TP: antlr_rust::parser_rule_context::ParserRuleContext<'input>,
+    FA: FnMut(&TC, &mut context::Context) -> diagnostic::Result<TR>,
+    TC: antlr_rust::parser_rule_context::ParserRuleContext<'input, TF = TP::TF, Ctx = TP::Ctx>
+        + 'input,
+{
+    parent
+        .children_of_type::<TC>()
+        .into_iter()
+        .enumerate()
+        .map(|(index, child)| {
+            push_child(
+                context,
+                &AntlrContextWrapper(child.as_ref()),
+                path::PathElement::Repeated(field.to_string(), index),
+                false,
+                |x: &AntlrContextWrapper<TC>, y| analyzer(x.0, y),
+            )
+        })
+        .unzip()
 }
